"""
Deploy hosts

hosts is a ordered dict,
key is the `c.host`
value is
```
{
    'host': '123.123.123.123',
    'user': 'root',
    'port': '22',
    'password': 'your password',
    'envs': ['test', 'prod'],
    'exclude_components': ['redis'],
}
```

"""
import inspect
import os
from collections import OrderedDict

from fabric import Connection

from .. import devops
from ..utils import say

hosts = OrderedDict()  # all hosts loads in top of fabric file

ENV = devops.ENV


def loads(host, value):
    hosts[host] = value

    # update project name
    previous_frame = inspect.currentframe().f_back
    filename, line_number, function_name, lines, index = inspect.getframeinfo(previous_frame)

    devops.PROJECT = os.path.dirname(filename).rsplit(os.sep, 1).pop(1).replace('-server', '')

    say(f'Project `{devops.PROJECT}` loads host successfully!', icon='✨', wrap='C')


def get_host_value(c):
    """
    get host orders:
        1. Provided in c.host
        2. Using the ENV name
        3. Using fixed `default`

    """
    try:
        host = c.host
    except (KeyError, AttributeError):
        if ENV in hosts:
            host = ENV
        else:
            host = 'default'

    if host not in hosts:
        raise ValueError(f'host ({host}) is not loaded yet!')

    host_value = hosts.get(host)

    # Check whether ENV and host envs match
    host_envs = host_value.get('envs', [])
    if ENV not in host_value.get('envs', []):
        raise ValueError(f'Runtime env ({ENV}) is not match host envs ({host_envs})')

    return host, host_value


def connect(c):
    # create a fabric connection
    host, host_value = get_host_value(c)
    default_user = 'root'
    return Connection(
        host_value.get('host'),
        user=host_value.get('user', default_user),
        connect_kwargs={
            'password': host_value.get('password'),
        },
    )



"""
Fabric tasks

Default environments is `dev`, `test`, `prod`

"""
from fabric import task

from .tasks import Hi, Environment, SetupServer, Deploy


@task
def hi(c):
    return Hi(c)()


@task
def dev(c):
    Environment(c).set_env('dev')


@task
def test(c):
    Environment(c).set_env('test')


@task
def prod(c):
    Environment(c).set_env('prod')


@task
def setup_server(c):
    """Setup a new CentOS server"""
    return SetupServer(c)()


@task
def deploy(c, requirements_update=False):
    """Deploy"""
    return Deploy(c)(requirements_update=requirements_update)



"""
Setup server configurations from a new server
Support CentOS 7
"""
from ..hosts import hosts, get_host_value, connect
from ...utils import say


class SetupServer:
    def __init__(self, c):
        self.c = connect(c)
        self.host_value = get_host_value(c)[1]

    def __call__(self, *args, **kwargs):
        self.setup_python3()
        self.check_components('nginx') and self.setup_nginx()
        self.check_components('supervisor') and self.setup_supervisor()
        self.check_components('mysql') and self.setup_mysql()
        self.check_components('redis') and self.setup_redis()

    def check_components(self, component):
        return component not in self.host_value.get('exclude_components', [])

    def setup_python3(self):
        """Install python3 and uWSGI"""
        # Install Python
        self.c.run('yum -y update')
        self.c.run('yum groupinstall "Development tools"')
        try:
            self.c.run('yum -y install wget gcc make zlib-devel')
        except Exception as ex:
            print('ex', ex)
        self.c.run('wget https://www.python.org/ftp/python/3.6.9/Python-3.6.9.tgz')
        self.c.run('tar xzf Python-3.6.9.tgz')
        self.c.run(
            'cd Python-3.6.9 && ./configure --with-ssl --prefix=/usr/local && make altinstall'
        )
        self.c.run('ln -s /usr/local/bin/python3.6 /usr/bin/python3')
        self.c.run('python3 -V')
        self.c.run('rm -rf Python-3.6.9')

        # Install uWSGI
        pypi_mirror_suffix = ' -i http://pypi.douban.com/simple/ --trusted-host pypi.douban.com'
        self.c.run(f'python3 -m pip install uWSGI==2.0.18 {pypi_mirror_suffix}')

    def setup_nginx(self):
        # Install Nginx
        self.c.run('yum install -y nginx')

    def setup_supervisor(self):
        # Install Supervisor in Python 2
        self.c.run('curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py')
        self.c.run('python get-pip.py')
        self.c.run('python -m pip install supervisor==4.1.0')

    def setup_mysql(self):
        download_url = 'https://dev.mysql.com/get/mysql80-community-release-el7-1.noarch.rpm'
        self.c.run(f'sudo rpm -Uvh {download_url}')
        self.c.run('sudo yum --enablerepo=mysql80-community install mysql-community-server')

    def setup_redis(self):
        version = '4.0.11'
        download_url = f'http://download.redis.io/releases/redis-{version}.tar.gz'
        self.c.run(f'curl -o redis-{version}.tar.gz {download_url}')
        self.c.run(f'tar -zxvf redis-{version}.tar.gz')

        # TODO: check the brew command
        self.c.run(f'sudo cd redis-{version} & make')
        self.c.run(f'cd src && make install PREFIX=/usr/local/redis')

        self.c.run(f'mkdir -p /usr/local/redis/conf')
        self.c.run(f'cp redis-{version}/redis.conf /usr/local/redis/conf')

        # start redis
        self.c.run('cd /usr/local/redis && bin/redis-server ./conf/redis.conf')



"""project managements

Using the commands to initialize new project or management existing project
"""

import os
import shutil
import subprocess

from ..utils import say, copy_template_to_file
from ... import core


def add_tags_to_base_settings(file_path):
    with open(file_path, 'r+') as f:
        content = f.read()

        # AUTH_USER_MODEL tag
        content = content.replace(
            '# Password validation', '\n'.join([
                '# {!AUTH_USER_MODEL}',
                '',
                '# Password validation',
            ])
        )

        f.seek(0)
        f.truncate()
        f.write(content)


def init(params):
    """Initialize django rework project"""

    project = params[0]
    try:
        path = params[1]
    except IndexError:
        path = None

    base_dir = os.getcwd()

    # when path is current dir, project dir is base dir
    if path == '.':
        project_dir = base_dir
    else:
        project_dir = os.path.join(base_dir, project)

    say(f'Initialing project: ``{project}`` using `django-admin` command')
    result = subprocess.run(["django-admin", "startproject", *params])

    if result.returncode != 0:
        say(f'Initialized failed!', icon='🌶 ', wrap='C')
        return False

    # Changed the settings files to satisfy multi environments
    say(f'Changed the settings files to satisfy multi environments')

    settings_folder = os.path.join(project_dir, project)

    # template variables
    kwargs = {
        'django_rework_version': core.__version__,
        'project': project,
    }

    #  1. Make a package named `settings`
    settings_package_path = os.path.join(settings_folder, 'settings')
    base_settings_path = os.path.join(settings_package_path, 'base')
    os.makedirs(base_settings_path)

    #  2. Move origin settings.py to settings/base/__init__.py
    origin_settings_file = os.path.join(settings_folder, 'settings.py')
    target_settings_file = os.path.join(base_settings_path, '__init__.py')
    shutil.move(origin_settings_file, target_settings_file)

    # Added template tag to base/__init__.py
    add_tags_to_base_settings(target_settings_file)

    #  3. Added multi env setting files
    settings_tpl_path = 'project/settings/'
    envs = ['prod', 'test', 'dev', 'dev.local']
    for env in envs:
        copy_template_to_file(f'{settings_tpl_path}{env}.py', base_dir, **kwargs)

    # Update settings

    # fabric DevOps
    copy_template_to_file('fabfile.py', base_dir, **kwargs)

    # Others
    copy_template_to_file('.editorconfig', base_dir, **kwargs)
    copy_template_to_file('.gitignore', base_dir, **kwargs)
    copy_template_to_file('requirements.txt', base_dir, **kwargs)

    print(f'{os.linesep} 🎨 Initialized completely!')



from django.contrib.auth.models import AbstractUser, BaseUserManager
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class UserManager(BaseUserManager):
    """
    Email is required to create users by default, and
    the implementation here removes the email
    """
    use_in_migrations = True

    def _create_user(self, username, password, is_staff, **extra_fields):
        """
        Creates and saves a User with the given username and password.
        """
        now = timezone.now()
        if not username:
            raise ValueError('The given username must be set')
        user = self.model(
            username=username, is_staff=is_staff, is_active=True, date_joined=now, **extra_fields
        )
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, username, password=None, **extra_fields):
        return self._create_user(username, password, False, **extra_fields)

    def create_superuser(self, username, password, **extra_fields):
        return self._create_user(username, password, True, **extra_fields)


class EnhancedAbstractUser(AbstractUser):
    mobile = models.CharField(_('mobile'), max_length=16, default='', db_index=True, blank=True)
    nickname = models.CharField(_('nickname'), max_length=128, default='')
    avatar = models.CharField(_('avatar'), max_length=256, default='')

    objects = UserManager()

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')
        abstract = True



import os
from string import Template
from .say import *


def get_project_name():
    """Determined the project path

    # TODO: the function must be improved
    There will be only one folder in root directory,
    it's the project name
    """
    for path in os.listdir('.'):
        if os.path.isdir(path):
            print('Determined the project path is: ', path)
            return path


def get_project_path(project=None):
    """Determined the project path"""
    base_dir = os.getcwd()
    project = project or get_project_name()
    return os.path.join(base_dir, project)


def get_settings_path(project=None):
    """Determined the project path"""
    base_dir = os.getcwd()
    project = project or get_project_name()

    path = os.path.join(base_dir, project, project, 'settings')
    if os.path.exists(path):
        return path

    path = os.path.join(base_dir, project, 'settings')
    if os.path.exists(path):
        return path

    raise Exception('Determine project path failed!')


def copy_template_to_file(file_path, target_path, **kwargs):
    """
    copy from template to create project file

    template path is `${file_path}+.tpl`

    :params file_path: human readable file path
    :params full_file_path:
    """
    template_base_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'templates')
    template_path = os.path.join(template_base_path, f'{file_path}.tpl')

    with open(template_path) as f:
        content = f.read()

    t = Template(content)

    target = os.path.join(target_path, file_path.replace('project', kwargs.get('project')))
    with open(target, 'w') as f:
        f.write(t.substitute(**kwargs))



import os
import sys

from . import project, app, deploy, migrate
from ..utils import say

COMMANDS = {
    'init': project.init,
    'add': app.add,
    'deploy': deploy.DeployCommand(),
    'migrate': migrate.migrate
}


def execute_from_command_line(argv=None):
    argv = argv or sys.argv[:]
    say(f'Hello, Rework CLI, you argv is {argv}', icon='🏂', wrap='C')

    command = argv[1]

    if command in COMMANDS:
        COMMANDS.get(command)(argv[2:])
    else:
        say(f'Command not found!', icon='🌶️ ', wrap='C')



import os
import subprocess

def migrate(params):
    """ migrate db """    

    print('Start to migrate DB')

    base_dir = os.getcwd()
    manage_dir = None

    for root, dirs, files in os.walk(base_dir):
        if os.path.exists(os.path.join(root, 'manage.py')):
            manage_dir = os.path.join(root, 'manage.py')
            break 

    if manage_dir:
        result = subprocess.run(['python3', manage_dir, 'migrate'])
        if result.returncode != 0:
            print('Migrate DB failed!')
        else:
            print('Migrate DB successed!')
    else:
        print('Not found manage.py!')


"""
Deploy to server 
"""
from ..hosts import get_host_value, connect
from ... import devops
from ...utils import say


class Deploy:
    def __init__(self, c):
        self.c = connect(c)
        self.host_value = get_host_value(c)[1]

        self.project = devops.PROJECT
        self.env = devops.ENV

        self.host = self.host_value.get('host')
        self.port = self.host_value.get('port', 22)
        self.password = self.host_value.get('password')

    def __call__(self, *args, **kwargs):
        if self.env == 'prod':
            root = f'/opt/projects/{self.project}-server/'
            service = f'{self.project}_server'
            branch = 'master'
        elif self.env == 'test':
            root = f'/opt/test-projects/{self.project}-server/'
            service = f'{self.project}_server_test'
            branch = 'test'
        elif self.env == 'dev':
            root = f'/opt/dev-projects/{self.project}-server/'
            service = f'{self.project}_server_dev'
            branch = 'dev'
        else:
            raise Exception('Un-supported Environment to deploy.')

        say(f'Project root is: {root}')

        self.pull(branch, root)
        if kwargs.get('requirements_update'):
            self.update_requirement(root)
        self.migrate(root)
        self.collect_static(root)
        self.copy_uwsgi(root)
        self.restart(service)

    def pull(self, branch, root):
        # Pull latest code
        say('Pull latest code from remote git')
        self.c.run(f'cd {root} && git checkout {branch} && git pull')

    def update_requirement(self, root):
        # Update requirements
        say('install requirements')
        self.c.run(f'cd {root} && python3 -m pip install -r requirements.txt')

    def migrate(self, root):
        # Models migrate
        say('Migrate database')
        self.c.run(
            f'cd {root} && python3 manage.py migrate --settings={self.project}.settings.{self.env}'
        )

    def collect_static(self, root):
        # Collect static
        say('Collect static')
        settings_suffix = f'--settings={self.project}.settings.{self.env}'
        self.c.run(f'cd {root} && python3 manage.py collectstatic {settings_suffix} --no-input')

    def copy_uwsgi(self, root):
        # Copy uwsgi files
        origin = f'{root}.deploy/supervisor/{self.project}_supervisor_{self.env}.conf'
        destination = f'/etc/supervisor/conf.d/'
        self.c.run(f'cp {origin} {destination}')

    def restart(self, service):
        # Restart uwsgi service
        say('Restart the uwsgi service')
        self.c.run(f'supervisorctl restart {service}')



import os
from abc import ABC


class BaseCommand(ABC):
    def __init__(self):
        self.base_dir = os.getcwd()
        self.core_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    def create_parser(self, prog_name, subcommand, **kwargs):
        pass



from ..hosts import hosts
from ...utils import say


class Hi:
    def __init__(self, c):
        self.c = c

    def __call__(self, *args, **kwargs):
        say('Hi, DevOps with Fabric is ready!', icon='🎉')
        say('loaded hosts: %s' % list(hosts.keys()), wrap='A')



from setuptools import find_packages, setup
from rework import core


def read(f):
    return open(f, 'r', encoding='utf-8').read()


# version Operators
# https://pip.pypa.io/en/latest/user_guide/#fixing-conflicting-dependencies
INSTALL_REQUIREMENTS = read('requirements.txt').splitlines()

setup(
    name='django-rework',
    version=core.__version__,
    description='Rapid develop framework base on Django',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/rework-union/django-rework',
    author='rework union',
    author_email='josh.yu_8@live.com',
    license='MIT',
    entry_points={
        'console_scripts': ['rework = rework.core.management:execute_from_command_line', ]
    },
    install_requires=INSTALL_REQUIREMENTS,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import os
import shutil

from ..utils.command import BaseCommand


class DeployCommand(BaseCommand):
    def handle(self, params):
        if params == ['--init']:
            try:
                working_dir = os.path.join(self.base_dir, 'deploy')
                core_dir = os.path.join(self.core_dir, 'deploy')
                shutil.copytree(core_dir, working_dir)
            except Exception as e:
                print(e)
        else:
            pass

    def __call__(self, params):
        return self.handle(params)



"""
Django Rework App managements
"""

import os
import re
import subprocess

from .. import utils
from ..utils import say


def setup_auth_user_model(content):
    # Setup `AUTH_USER_MODEL` to `basic.User` if adding app is `users`,
    # and generate `basic` App
    result = subprocess.run(["django-admin", "startapp", "basic"])
    if result.returncode != 0:
        say(f'Generate `basic` App failed!', icon='🌶 ', wrap='C')
        return False

    pattern = '# {!AUTH_USER_MODEL}'
    block = """AUTH_USER_MODEL = 'basic.User'"""
    content = content.replace(pattern, block)

    return content


def add(params):
    """Add apps to project"""
    app = params[0].lower()
    say(f'Start adding app: {app}')

    # Add app name to base settings
    settings_file = os.path.join(utils.get_settings_path(), 'base', '__init__.py')
    with open(settings_file, 'r+') as file:
        # check where app in INSTALLED_APPS
        content = file.read()
        print('settings content is', content)

        app_full_name = f'rework.contrib.{app}'
        installed_apps_pattern = r'INSTALLED_APPS\s*\=\s*\[[\s\S]*?\]'

        installed_apps_match = re.search(installed_apps_pattern, content)
        if not installed_apps_match:
            print('There is no `INSTALLED_APPS` block in your settings')
            return False

        installed_apps_block = installed_apps_match.group()

        for exist_app in re.findall(r"(?<=').*(?=')", installed_apps_block):
            if exist_app == app_full_name:
                say(f'[ERROR] App {app_full_name} is already exists')
                return False

        installed_apps_block = re.sub(
            r'\n]',
            f"\n    '{app_full_name}',\n]",
            installed_apps_block,
        )

        say(f'installed_apps_block {installed_apps_block}')

        content = re.sub(installed_apps_pattern, installed_apps_block, content)

        if app == 'users':
            content = setup_auth_user_model(content)

        say(f'content {content}')
        file.seek(0)
        file.truncate()
        file.write(content)

    say('Added completely!')



def say(content, icon=None, wrap=''):
    """convenient way to output message

    `wrap` choices is
        'A' - append wrap after content,
        'B' - before content,
        'C' - both
    """
    leading = f' {icon:3}' if icon else '  - '
    wrap = wrap.upper()
    if wrap in ['B', 'C']:
        print()

    print('{}{}'.format(leading, content))

    if wrap in ['A', 'C']:
        print()
